// biginteger.hpp

#ifndef BIGINTEGER_HPP
#define BIGINTEGER_HPP

#include <algorithm>
#include <iostream>
#include <sstream> 
#include <iomanip>
#include <cstdint>
#include <cstddef>
#include <cstring>
#include <string>
#include <vector>
#include <cmath>

extern  bool  fftOK;
extern double fftSquareWorstError;

class BigInteger {

public:
    ~BigInteger();
    BigInteger();
    BigInteger(int32_t a);
    BigInteger(int64_t a);
    BigInteger(uint32_t a);
    BigInteger(uint64_t a);
    BigInteger(const std::string& a);
    BigInteger(const char* a);
    BigInteger(const BigInteger& a);
    int32_t nBits() const;
    bool getBit(int32_t n) const;
    void setBit(int32_t n, bool v);
    int32_t getSize() const;
    int32_t getElemt(int32_t n) const;
    void setElemt(int32_t n, int32_t e);
    void popElemt();
    void finbi();
    std::vector<int32_t> getElems() const;
    void setElems(std::vector<int32_t> v);
    int16_t getSigne() const;
    void chgSigne();
    void setbi0();
    void setbi1();
    BigInteger operator -  () const;
    BigInteger operator =  (BigInteger a);
    BigInteger operator += (const BigInteger& a);
    BigInteger operator -= (const BigInteger& a);
    BigInteger operator *= (const BigInteger& a);
    BigInteger operator /= (const BigInteger& a);
    BigInteger operator %= (const BigInteger& a);
    BigInteger operator ^= (const BigInteger& a);
    BigInteger operator ++ ();
    BigInteger operator ++ (int);
    BigInteger operator -- ();
    BigInteger operator -- (int);

private:
    // Le BigInteger est reprsent par _signe et _elems :
    // _elems[0] + _elems[1]*basebi + _elems[2]*basebi^2 + ...
    // _elems[i] va de 0  basebi-1 = (2^bitsbi)-1
    std::vector<int32_t> _elems;
    // signe = -1 ou +1 ( le signe de 0 est +1 )
    int16_t _signe;
};

BigInteger operator + (const BigInteger& l, const BigInteger& r);
BigInteger operator - (const BigInteger& l, const BigInteger& r);
BigInteger operator * (const BigInteger& l, const BigInteger& r);
BigInteger operator / (const BigInteger& l, const BigInteger& r);
BigInteger operator % (const BigInteger& l, const BigInteger& r);
BigInteger operator ^ (const BigInteger& x, const BigInteger& n);
bool operator == (const BigInteger& l, const BigInteger& r);
bool operator != (const BigInteger& l, const BigInteger& r);
bool operator >  (const BigInteger& l, const BigInteger& r);
bool operator >= (const BigInteger& l, const BigInteger& r);
bool operator <  (const BigInteger& l, const BigInteger& r);
bool operator <= (const BigInteger& l, const BigInteger& r);
std::ostream& operator << (std::ostream& ost, const BigInteger& a);
std::istream& operator >> (std::istream& ist, BigInteger& a);
int32_t cmpbi(const BigInteger& a, const BigInteger& b);
BigInteger addbi(const BigInteger& a, const BigInteger& b);
BigInteger subbi(const BigInteger& a, const BigInteger& b);
BigInteger mulbt(const BigInteger& a, const BigInteger& b);
BigInteger mulbk(const BigInteger& a, const BigInteger& b);
BigInteger mulbi(const BigInteger& a, const BigInteger& b);
BigInteger modbi(const BigInteger& a, const BigInteger& n);
BigInteger expbi(const BigInteger& a, const BigInteger& n);
void divbi(BigInteger* q, BigInteger* r, const BigInteger& n, const BigInteger& d);
void sqrbi(BigInteger* q, BigInteger* r, const BigInteger& n);
BigInteger absbi(BigInteger n);
long double logbi(BigInteger n);
BigInteger factbi(const BigInteger& n);
BigInteger primbi(const BigInteger& n);
BigInteger ppdpbi(const BigInteger& n);
BigInteger ppnpbi(const BigInteger& n);
BigInteger pgcdbi(const BigInteger& a, const BigInteger& b);
BigInteger ppcmbi(const BigInteger& a, const BigInteger& b);
BigInteger randbi(const BigInteger& a, const BigInteger& b);
BigInteger invmbi(const BigInteger& a, const BigInteger& m);
BigInteger expmbi(BigInteger n, BigInteger p, BigInteger m);
bool isnegabi(const BigInteger& n);
bool ispairbi(const BigInteger& n);
bool isprembi(const BigInteger& n);
BigInteger ndivbi(const BigInteger& n);
std::vector<BigInteger> fctpbi(const BigInteger& n);
std::string bi2string(BigInteger x, int16_t radix);
std::string bi2deci(const BigInteger& x);
std::string bi2hexa(const BigInteger& x);
int64_t bi2int(const BigInteger& a);

namespace {
    // bitsbi maximum = 30 (_elems[i] est un int32_t)
    const int32_t bitsbi = 20;
    // basebi = 2^bitsbi
    const int32_t basebi = 1 << bitsbi;
    const BigInteger binul = BigInteger(0);
    const BigInteger bione = BigInteger(1);
}

#endif  // BIGINTEGER_HPP